'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var parser = require('@babel/parser');
var types = require('@babel/types');
var languageLibrary = require('@webdoc/language-library');
var traverse = require('@babel/traverse');
var model = require('@webdoc/model');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () { return e[k]; }
        });
      }
    });
  }
  n["default"] = e;
  return Object.freeze(n);
}

var parser__namespace = /*#__PURE__*/_interopNamespace(parser);
var traverse__default = /*#__PURE__*/_interopDefaultLegacy(traverse);

const VARIABLE = 0xff;
const PARAMETER = 0xffff;
const variableRegistry = [];
let registryBlocks = 0;
function declareParameter(id) {
  variableRegistry[registryBlocks - 1][id] = PARAMETER;
}
function declareVariable(id) {
  variableRegistry[registryBlocks - 2][id] = VARIABLE;
}
function createBlock() {
  ++registryBlocks;

  if (variableRegistry.length < registryBlocks) {
    let i = variableRegistry.length;
    variableRegistry.length = registryBlocks;

    for (; i < registryBlocks; i++) {
      variableRegistry[i] = {};
    }
  } else {
    variableRegistry[registryBlocks - 1] = {};
  }
}
function removeBlock() {
  --registryBlocks;
}
function clearRegistry() {
  registryBlocks = 0;
}
function queryType(id) {
  for (let i = registryBlocks - 1; i >= 0; i--) {
    const block = variableRegistry[i];

    for (const vid in block) {
      if (vid === id) {
        return block[id];
      }
    }
  }

  return null;
}

function extract(node) {
  const doc = node.value;
  node.documented = true;

  if (!doc) {
    return;
  }

  if (!doc.startsWith("*") && !doc.startsWith("!")) {
    return;
  }

  const docLines = doc.split("\n");

  for (let i = 0; i < docLines.length; i++) {
    docLines[i] = docLines[i].trimEnd();

    if (docLines[i].trimStart().startsWith("*")) {
      docLines[i] = docLines[i].trimStart().replace("*", "");
    }
  }

  return docLines.join("\n");
}

const mode = {
  current: "typescript"
};
function extractExtends(node) {
  if (types.isClassDeclaration(node) || types.isClassExpression(node)) {
    if (types.isIdentifier(node.superClass)) {
      return [node.superClass.name];
    }

    if (types.isMemberExpression(node.superClass)) {
      return [reduceMemberExpression(node.superClass)];
    }

    return null;
  }

  if (!node.extends) {
    return;
  }

  return node.extends.map(exs => {
    if (types.isIdentifier(exs.id)) {
      return exs.id.name;
    }

    if (types.isQualifiedTypeIdentifier(exs.id)) {
      return reduceQualifiedName(exs.id);
    }

    if (types.isIdentifier(exs.expression)) {
      return exs.expression.name;
    }

    if (types.isTSQualifiedName(exs.expression)) {
      return reduceTSQualifiedName(exs.expression);
    }

    return "";
  });
}
function extractImplements(node) {
  return node.implements ? node.implements.map(impls => {
    if (types.isClassImplements(impls)) {
      return impls.id.name;
    }

    if (types.isTSExpressionWithTypeArguments(impls)) {
      if (types.isIdentifier(impls.expression)) {
        return impls.expression.name;
      }

      if (types.isQualifiedTypeIdentifier(impls.id)) {
        return reduceQualifiedName(impls.id);
      }

      if (types.isIdentifier(impls.expression)) {
        return impls.expression.name;
      }

      if (types.isTSQualifiedName(impls.expression)) {
        return reduceTSQualifiedName(impls.expression);
      }

      return "";
    }
  }) : null;
}
function extractParams(node) {
  if (!node.params && !node.parameters) {
    return [];
  }

  const nodeParams = node.params ? node.params : node.parameters;
  const params = [];

  for (let i = 0; i < nodeParams.length; i++) {
    let paramNode = nodeParams[i];
    let param;
    let paramTypeAnnotation;

    if (types.isTSParameterProperty(paramNode)) {
      paramNode = paramNode.parameter;
    }

    if (types.isIdentifier(paramNode)) {
      param = {
        identifier: paramNode.name,
        optional: paramNode.optional || false
      };
    } else if (types.isRestElement(paramNode)) {
      const argument = paramNode.argument;
      param = {
        identifier: types.isIdentifier(argument) ? argument.name : "rest_args",
        optional: paramNode.optional || false,
        variadic: true
      };
    } else if (types.isAssignmentPattern(paramNode)) {
      const extraRaw = paramNode.right.extra && paramNode.right.extra.raw;
      const [defaultValue, dataType] = extractAssignedValue(paramNode.right);

      if (types.isObjectPattern(paramNode.left)) {
        try {
          params.push(({
            identifier: `__arg${i}`,
            optional: false,
            ...(paramNode.left.typeAnnotation && {
              dataType: extractType(paramNode.left.typeAnnotation)
            })
          }));
          params.push(...extractParamsFromObjectPatternRecursive(paramNode.left, paramNode.right, paramNode.left.typeAnnotation ? paramNode.left.typeAnnotation.typeAnnotation : null));
        } catch (e) {
          (params).flawed = true;
          languageLibrary.log.error(languageLibrary.tag.Indexer, `Failed parse param: ${e}`);
        }
      } else {
        param = {
          identifier: paramNode.left.name,
          optional: paramNode.optional || false,
          default: paramNode.right.raw || extraRaw || defaultValue,
          dataType
        };
        paramTypeAnnotation = paramNode.left.typeAnnotation;
      }
    } else if (types.isObjectPattern(paramNode)) {
      try {
        params.push(({
          identifier: `__arg${i}`,
          optional: false,
          ...(paramNode.typeAnnotation && {
            dataType: extractType(paramNode.typeAnnotation)
          })
        }));
        params.push(...extractParamsFromObjectPatternRecursive(paramNode, null, paramNode.typeAnnotation ? paramNode.typeAnnotation.typeAnnotation : null));
      } catch (e) {
        (params).flawed = true;
        languageLibrary.log.error(languageLibrary.tag.Indexer, `Failed parse param: ${e}`);
      }
    } else {
      (params).flawed = true;
      languageLibrary.log.error(languageLibrary.tag.Indexer, "Parameter node couldn't be parsed, " + "enable warning logs for more detail");
      languageLibrary.log.warn(languageLibrary.tag.Indexer, JSON.stringify(paramNode, null, 2));
    }

    if (param && paramTypeAnnotation) {
      param.dataType = extractType(paramTypeAnnotation);
    } else if (param && paramNode.typeAnnotation) {
      param.dataType = extractType(paramNode.typeAnnotation);
    }

    if (param) {
      params.push(param);
    }
  }

  return params;
}

function extractParamsFromObjectPatternRecursive(left, right, typeAnnotation) {
  const params = [];

  for (const prop of left.properties) {
    if (types.isObjectProperty(prop)) {
      let defaultValue;

      if (types.isAssignmentPattern(prop.value)) {
        defaultValue = prop.value.right;
      } else if (right) {
        for (const defaultValueProp of right.properties) {
          if (defaultValueProp.key && defaultValueProp.key.name === prop.key.name) {
            defaultValue = defaultValueProp.value;
            break;
          }
        }
      }

      let valueTypeAnnotation;

      if (typeAnnotation && types.isObjectTypeAnnotation(typeAnnotation)) {
        for (const typeProp of typeAnnotation.properties) {
          if (types.isIdentifier(typeProp.key) && typeProp.key.name === prop.key.name) {
            valueTypeAnnotation = {
              typeAnnotation: typeProp.value
            };
            break;
          }
        }
      } else if (typeAnnotation && types.isTSTypeLiteral(typeAnnotation)) {
        for (const typeProp of typeAnnotation.members) {
          if (types.isIdentifier(typeProp.key) && typeProp.key.name === prop.key.name) {
            valueTypeAnnotation = typeProp.typeAnnotation;
            break;
          }
        }
      }

      if (types.isObjectPattern(prop.value)) {
        const embeddedParams = extractParamsFromObjectPatternRecursive(prop.value, defaultValue, valueTypeAnnotation ? valueTypeAnnotation.typeAnnotation : null);
        const prefix = types.isIdentifier(prop.key) ? prop.key.name : "$error$";

        for (const embeddedParam of embeddedParams) {
          embeddedParam.identifier = `.${prefix}${embeddedParam.identifier}`;
        }

        params.push(...embeddedParams);
      } else if (types.isIdentifier(prop.key)) {
        const [defaultValueRaw, impliedDataType] = defaultValue ? extractAssignedValue(defaultValue) : [null, null];
        const param = {
          identifier: "." + prop.key.name,
          optional: valueTypeAnnotation && valueTypeAnnotation.optional,
          variadic: false,
          ...(defaultValueRaw && {
            default: defaultValueRaw
          }),
          dataType: valueTypeAnnotation ? extractType(valueTypeAnnotation) : impliedDataType
        };
        params.push(param);
      }
    }
  }

  return params;
}

function extractReturns(node) {
  if (node.returnType) {
    return [{
      dataType: extractTypeFailsafe(node.returnType)
    }];
  }

  if (node.typeAnnotation) {
    return [{
      dataType: extractTypeFailsafe(node.typeAnnotation)
    }];
  }

  return [];
}
function extractRawExpression(expression) {
  if (types.isLiteral(expression)) {
    if (types.isStringLiteral(expression)) {
      return `"${expression.value}"`;
    } else {
      return `${expression.value}`;
    }
  }

  if (types.isUnaryExpression(expression)) {
    return `-${extractRawExpression(expression.argument) || ""}`;
  }
}
function extractAssignedValue(node) {
  let defaultValue;
  let dataType;

  if (types.isLiteral(node)) {
    if (types.isStringLiteral(node)) {
      defaultValue = `"${node.value}"`;
      dataType = model.createSimpleKeywordType("string");
    } else {
      defaultValue = `${node.value}`;

      if (types.isNumericLiteral(node)) {
        dataType = model.createSimpleKeywordType("number");
      } else if (types.isBooleanLiteral(node)) {
        dataType = model.createSimpleKeywordType("boolean");
      }
    }
  } else if (types.isExpression(node)) {
    defaultValue = extractRawExpression(node);

    if (typeof defaultValue === "string") {
      if (!isNaN(parseFloat(defaultValue))) {
        dataType = model.createSimpleKeywordType("number");
      }
    }
  }

  return [defaultValue, dataType];
}
function extractTypeFailsafe(node) {
  if (node.typeAnnotation) {
    if (types.isFlowType(node.typeAnnotation) || mode.current === "flow") {
      return resolveFlowDataType(node.typeAnnotation);
    }

    return resolveTSDataType(node.typeAnnotation);
  }

  console.error(node);
  console.error("Failsafe type extraction found invalid type.");
  return model.createSimpleKeywordType("invalid");
}
function extractType(node) {
  if (node.typeAnnotation) {
    if (types.isFlowType(node.typeAnnotation) || mode.current === "flow") {
      return resolveFlowDataType(node.typeAnnotation);
    }

    return resolveTSDataType(node.typeAnnotation);
  }
}

function reduceMemberExpression(expression) {
  let longname = "";

  if (types.isTSAsExpression(expression)) {
    expression = expression.expression;
  }

  if (types.isThisExpression(expression)) {
    return "this";
  }

  while (expression.object) {
    longname = expression.property.name + (longname ? "." + longname : "");
    expression = expression.object;
  }

  longname = (types.isThisExpression(expression) ? "this" : expression.name) + (longname ? "." : "") + longname;
  return longname;
}

function reduceQualifiedName(type) {
  let name = "";

  while (type && !types.isIdentifier(type)) {
    name = type.id.name + (name ? "." : "") + name;
    type = type.qualification;
  }

  if (type) {
    name = type.name + "." + name;
  }

  return name;
}

function reduceTSQualifiedName(type) {
  let name = "";

  while (type && !types.isIdentifier(type)) {
    name = type.right.name + (name ? "." : "") + name;
    type = type.left;
  }

  if (type) {
    name = type.name + "." + name;
  }

  return name;
}

function resolveFlowDataType(type) {
  if (types.isTypeAnnotation(type)) {
    return resolveFlowDataType(type.typeAnnotation);
  }

  if (types.isIdentifier(type)) {
    return model.createSimpleDocumentedType(type.name);
  }

  if (types.isFunctionTypeParam(type)) {
    const name = type.name ? type.name.name : "";
    const dataType = model.cloneType(resolveFlowDataType(type.typeAnnotation));
    dataType[0] = name + ": " + dataType[0];
    dataType.template = name + ": " + dataType.template;
    return dataType;
  }

  if (types.isGenericTypeAnnotation(type)) {
    let dataType = resolveFlowDataType(type.id);

    if (type.typeParameters) {
      const typeParamsDataType = resolveFlowDataType(type.typeParameters);
      console.log(typeParamsDataType);

      if (typeParamsDataType) {
        dataType = model.createComplexType("", dataType, typeParamsDataType);
      }
    }

    return dataType;
  }

  if (types.isTypeParameterInstantiation(type)) {
    const dataType = model.createComplexType(", ", ...type.params.map(param => resolveFlowDataType(param)));
    dataType[0] = `<${dataType[0]}>`;
    dataType.template = `<${dataType.template}>`;
    return dataType;
  }

  if (types.isObjectTypeProperty(type)) {
    const key = types.isStringLiteral(type) ? `"${type.key.value}"` : type.key.name;
    const dataType = resolveFlowDataType(type.value);
    const prefix = type.kind !== "init" ? `${type.kind} ` : "";
    const suffix = type.kind !== "init" ? "()" : "";
    const separator = type.optional ? "?:" : ":";
    dataType[0] = `${prefix}${key}${suffix}${separator} ${dataType[0]}`;
    dataType.template = `${prefix}${key}${suffix}${separator} ${dataType.template}`;
    return dataType;
  }

  if (types.isQualifiedTypeIdentifier(type)) {
    return model.createSimpleDocumentedType(reduceQualifiedName(type));
  }

  if (types.isObjectTypeIndexer(type)) {
    const id = type.id ? type.id.name : null;
    const key = resolveFlowDataType(type.key);
    const value = resolveFlowDataType(type.value);

    if (id !== null) {
      key[0] = `[${id}: ${key[0]}]`;
      key.template = `[${id} : ${key.template}]`;
    } else {
      key[0] = `[${key[0]}]`;
      key.template = `[${key.template}]`;
    }

    return model.createComplexType(": ", key, value);
  }

  if (types.isFlowType(type)) {
    if (types.isAnyTypeAnnotation(type)) {
      return model.createSimpleKeywordType("any");
    }

    if (types.isArrayTypeAnnotation(type)) {
      const dataType = resolveFlowDataType(type.elementType);
      dataType[0] += "[]";
      dataType.template += "[]";
      return dataType;
    }

    if (types.isBooleanTypeAnnotation(type)) {
      return model.createSimpleKeywordType("boolean");
    }

    if (types.isBooleanLiteralTypeAnnotation(type)) {
      return model.createSimpleKeywordType(`${type.value}`);
    }

    if (types.isNullLiteralTypeAnnotation(type)) {
      return model.createSimpleKeywordType("null");
    }

    if (types.isFunctionTypeAnnotation(type)) {
      const params = type.params.map(param => resolveFlowDataType(param));

      if (type.rest) {
        params.push(model.createRestDataType(resolveFlowDataType(type.rest)));
      }

      return model.createFunctionType(params, type.returnType ? resolveFlowDataType(type.returnType) : null);
    }

    if (types.isIntersectionTypeAnnotation(type)) {
      return model.createComplexType(" & ", type.types.map(type => resolveFlowDataType(type)));
    }

    if (types.isMixedTypeAnnotation(type)) {
      return model.createSimpleKeywordType("mixed");
    }

    if (types.isNullableTypeAnnotation(type)) {
      const dataType = resolveFlowDataType(type.typeAnnotation);
      dataType[0] = `?${dataType[0]}`;
      dataType.template = `?${dataType.template}`;
      return dataType;
    }

    if (types.isNumberLiteralTypeAnnotation(type)) {
      return model.createSimpleKeywordType(`${type.value}`);
    }

    if (types.isNumberTypeAnnotation(type)) {
      return model.createSimpleKeywordType("number");
    }

    if (types.isObjectTypeAnnotation(type)) {
      const properties = (type.properties || []).map(prop => resolveFlowDataType(prop));
      const indexers = (type.indexers || []).map(indexer => resolveFlowDataType(indexer));
      const dataType = model.createComplexType(", ", ...properties, ...indexers);
      dataType[0] = `{ ${dataType[0]} }`;
      dataType.template = `{ ${dataType.template} }`;
      return dataType;
    }

    if (types.isStringLiteralTypeAnnotation(type)) {
      return model.createSimpleKeywordType(`"${type.value}"`);
    }

    if (types.isStringTypeAnnotation(type)) {
      return model.createSimpleKeywordType("string");
    }

    if (types.isThisTypeAnnotation(type)) {
      return model.createSimpleKeywordType("this");
    }

    if (types.isTupleTypeAnnotation(type)) {
      const dataType = model.createComplexType(", ", ...type.types.map(t => resolveFlowDataType(t)));
      dataType[0] = `[${dataType[0]}]`;
      dataType.template = `[${dataType.template}]`;
      return dataType;
    }

    if (types.isTypeofTypeAnnotation(type)) {
      const dataType = resolveFlowDataType(type);
      dataType[0] = `typeof ${dataType[0]}`;
      dataType.template = `typeof ${dataType.template}`;
      return dataType;
    }

    if (types.isUnionTypeAnnotation(type)) {
      return model.createComplexType(" | ", ...type.types.map(t => resolveFlowDataType(t)));
    }

    if (types.isVoidTypeAnnotation(type)) {
      return model.createSimpleKeywordType("void");
    }
  }

  console.log(type);
  console.log("***unknown (flow) data-type node***");
  return model.createSimpleKeywordType("unknown");
}

function resolveTSDataType(type) {
  if (types.isTypeAnnotation(type) || types.isTSTypeAnnotation(type)) {
    return resolveTSDataType(type.typeAnnotation);
  }

  if (types.isIdentifier(type) && type.typeAnnotation) {
    const dataType = model.cloneType(resolveTSDataType(type.typeAnnotation));
    dataType[0] = type.name + ": " + dataType[0];
    dataType.template = type.name + ": " + dataType.template;
    return dataType;
  }

  if (types.isIdentifier(type)) {
    return model.createSimpleDocumentedType(type.name);
  } else if (types.isTSQualifiedName(type)) {
    return model.createSimpleDocumentedType(reduceTSQualifiedName(type));
  }

  if (types.isTSType(type)) {
    if (types.isTSTypeReference(type)) {
      let dataType = resolveTSDataType(type.typeName);

      if (type.typeParameters) {
        const typeParamsDataType = resolveTSDataType(type.typeParameters);

        if (typeParamsDataType) {
          dataType = model.createComplexType("", dataType, typeParamsDataType);
        }
      }

      return dataType;
    }

    if (types.isTSAnyKeyword(type)) {
      return model.createSimpleKeywordType("any");
    }

    if (types.isTSArrayType(type)) {
      const dataType = resolveTSDataType(type.elementType);
      dataType[0] += "[]";
      dataType.template += "[]";
      return dataType;
    }

    if (types.isTSBooleanKeyword(type)) {
      return model.createSimpleKeywordType("boolean");
    }

    if (types.isTSBigIntKeyword(type)) {
      return model.createSimpleKeywordType("BigInt");
    }

    if (types.isTSNeverKeyword(type)) {
      return model.createSimpleKeywordType("never");
    }

    if (types.isTSNullKeyword(type)) {
      return model.createSimpleKeywordType("null");
    }

    if (types.isTSNumberKeyword(type)) {
      return model.createSimpleKeywordType("number");
    }

    if (types.isTSObjectKeyword(type)) {
      return model.createSimpleKeywordType("object");
    }

    if (types.isTSStringKeyword(type)) {
      return model.createSimpleKeywordType("string");
    }

    if (types.isTSSymbolKeyword(type)) {
      return model.createSimpleKeywordType("Symbol");
    }

    if (types.isTSUndefinedKeyword(type)) {
      return model.createSimpleKeywordType("undefined");
    }

    if (types.isTSUnknownKeyword(type)) {
      return model.createSimpleKeywordType("unknown");
    }

    if (types.isTSVoidKeyword(type)) {
      return model.createSimpleKeywordType("void");
    }

    if (types.isTSThisType(type)) {
      return model.createSimpleDocumentedType("this");
    }

    if (types.isTSLiteralType(type)) {
      if (types.isStringLiteral(type.literal)) {
        return model.createSimpleKeywordType(`"${type.literal.value}"`);
      }

      return model.createSimpleKeywordType(type.literal.value);
    }

    if (types.isTSFunctionType(type) || types.isTSConstructorType(type)) {
      return model.createFunctionType(type.parameters.map(param => resolveTSDataType(param)), type.typeAnnotation ? resolveTSDataType(type.typeAnnotation) : null);
    }

    if (types.isTSRestType(type)) {
      return model.createRestDataType(resolveTSDataType(type.typeAnnotation));
    }

    if (types.isTSUnionType(type)) {
      return model.createComplexType(" | ", ...type.types.map(subtype => resolveTSDataType(subtype)));
    }

    if (types.isTSIntersectionType(type)) {
      return model.createComplexType(" & ", ...type.types.map(subtype => resolveTSDataType(subtype)));
    }

    if (types.isTSTupleType(type)) {
      const dataType = model.createComplexType(", ", ...type.elementTypes.map(subtype => resolveTSDataType(subtype)));
      dataType[0] = `[${dataType[0]}]`;
      dataType.template = `[${dataType.template}]`;
      return dataType;
    }

    if (types.isTSParenthesizedType(type)) {
      const dataType = model.cloneType(resolveTSDataType(type.typeAnnotation));
      dataType[0] = `(${dataType[0]})`;
      dataType.template = `(${dataType.template})`;
      return dataType;
    }

    if (types.isTSTypePredicate(type)) {
      const dataType = model.cloneType(resolveTSDataType(type.typeAnnotation));
      dataType[0] = `${type.parameterName.name} is ${dataType[0]}`;
      dataType.template = `${type.parameterName.name} is ${dataType.template}`;
      return dataType;
    }

    if (types.isTSTypeQuery(type)) {
      const dataType = resolveTSDataType(type.exprName);
      dataType[0] = `typeof ${dataType[0]}`;
      dataType.template = `typeof ${dataType.template}`;
      return dataType;
    }

    if (types.isTSMappedType(type)) {
      const dataType = resolveTSDataType(type.typeAnnotation);
      const name = type.typeParameter.name;
      const operator = type.typeParameter.constraint.operator;
      const constraintType = resolveTSDataType(type.typeParameter.constraint);
      const attribs = `${type.readonly ? "readonly " : ""}${type.static ? "static " : ""}`;
      constraintType[0] = `{ ${attribs}[${name} in ${operator} ${constraintType[0]}] }`;
      constraintType.template = `{ ${attribs}[${name} in ${operator} ${constraintType.template}] }`;
      return model.createComplexType(": ", constraintType, dataType);
    }
  }

  if (types.isTSTypeLiteral(type)) {
    const dataType = model.createComplexType(", ", ...type.members.map(mem => resolveTSDataType(mem)));
    dataType[0] = "{ " + dataType[0] + " }";
    dataType.template = `{ ${dataType.template} }`;
    return dataType;
  }

  if (types.isTSIndexedAccessType(type)) {
    const index = resolveTSDataType(type.indexType);
    index[0] = `[${index[0]}]`;
    index.template = `[${index.template}]`;
    const object = resolveTSDataType(type.objectType);
    return model.createComplexType("", object, index);
  }

  if (types.isTSPropertySignature(type)) {
    const dataType = resolveTSDataType(type.typeAnnotation);
    const key = (type.key.name || `[${type.key.value}]`) + (type.optional ? "?" : "");
    const attribs = `${type.readonly ? "readonly " : ""}${type.static ? "static " : ""}`;
    dataType[0] = `${attribs}${key} : ${dataType[0]}`;
    dataType.template = `${attribs}${key} : ${dataType.template}`;
    return dataType;
  }

  if (types.isTSIndexSignature(type)) {
    const params = model.createComplexType(", ", ...type.parameters.map(param => {
      return resolveTSDataType(param);
    }));
    params[0] = `[${params[0]}]`;
    params.template = `[${params.template}]`;
    const dataType = model.createComplexType(": ", params, resolveTSDataType(type.typeAnnotation));
    return dataType;
  }

  if (types.isTSConstructSignatureDeclaration(type)) {
    const params = model.createComplexType(", ", ...type.parameters.map(param => resolveTSDataType(param)));
    params[0] = `new (${params[0]})`;
    params.template = `new (${params.template})`;

    if (type.typeAnnotation) {
      return model.createComplexType(": ", params, resolveTSDataType(type.typeAnnotation));
    } else {
      return params;
    }
  }

  if (types.isTSTypeOperator(type)) {
    if (type.operator) {
      return model.createSimpleKeywordType(type.operator);
    }
  }

  if (types.isTSTypeParameterInstantiation(type)) {
    const typeParams = model.createComplexType(", ", ...type.params.map(param => resolveTSDataType(param)));
    typeParams[0] = `<${typeParams[0]}>`;
    typeParams.template = `<${typeParams.template}>`;
    return typeParams;
  }

  if (types.isVoidTypeAnnotation(type)) {
    return model.createSimpleKeywordType("void");
  }

  if (types.isRestElement(type)) {
    const dataType = resolveTSDataType(type.typeAnnotation);
    dataType[0] = `...${type.argument.name}: ${dataType[0]}`;
    dataType.template = `...${type.argument.name}: ${dataType.template}`;
    return dataType;
  }

  console.log(type);
  console.log("***unknown (typescript) data-type node***");
  return model.createSimpleKeywordType("unknown");
}

function extractSymbol(nodePath, parent, out) {
  const node = nodePath.node;
  let name;
  let flags = out.flags || 0;
  let isInit = false;
  let initComment;
  let init;
  const nodeSymbol = out.nodeSymbol || {
    meta: {}
  };

  if (types.isClassMethod(node) && (node.kind === "method" || node.kind === "constructor") || types.isObjectMethod(node) || types.isTSDeclareMethod(node) && (node.kind === "method" || node.kind === "constructor")) {
    name = node.key.name;

    if (node.abstract) {
      nodeSymbol.meta.abstract = true;
    }

    nodeSymbol.meta.access = node.access || node.accessibility;
    nodeSymbol.meta.scope = node.static ? "static" : "instance";
    nodeSymbol.meta.type = "MethodDoc";
    nodeSymbol.meta.params = extractParams(node);
    nodeSymbol.meta.returns = extractReturns(node);
  } else if (types.isClassProperty(node) || types.isClassMethod(node) && (node.kind === "get" || node.kind === "set")) {
    name = node.key.name;
    nodeSymbol.meta.access = node.access || node.accessibility;
    nodeSymbol.meta.readonly = node.readonly;
    nodeSymbol.meta.dataType = types.isClassProperty(node) ? extractType(node) : types.isClassMethod(node) && node.returnType ? extractType(node.returnType) : null;
    nodeSymbol.meta.scope = node.static ? "static" : "instance";
    nodeSymbol.meta.type = "PropertyDoc";
    const [defaultValue, dataType] = extractAssignedValue(node.value);

    if (typeof defaultValue === "string") {
      nodeSymbol.meta.defaultValue = defaultValue;

      if (!nodeSymbol.meta.dataType) {
        nodeSymbol.meta.dataType = dataType;
      }
    }
  } else if (types.isTSParameterProperty(node)) {
    const parameter = node.parameter;

    if (types.isIdentifier(parameter)) {
      name = parameter.name;
    } else if (types.isAssignmentPattern(parameter)) {
      name = parameter.left.name;
      nodeSymbol.meta.defaultValue = parameter.right.raw || parameter.right.extra && parameter.right.extra.raw;
    }

    nodeSymbol.meta.access = node.accessibility;
    nodeSymbol.meta.object = "this";
    nodeSymbol.meta.readonly = node.readonly;
    nodeSymbol.meta.scope = "instance";
    nodeSymbol.meta.type = "PropertyDoc";

    if (parameter.typeAnnotation) {
      nodeSymbol.meta.dataType = extractType(parameter.typeAnnotation);
    }

    flags |= languageLibrary.OBLIGATE_LEAF;
  } else if (types.isClassDeclaration(node) || types.isClassExpression(node)) {
    name = node.id ? node.id.name : "";
    nodeSymbol.meta.extends = extractExtends(node);
    nodeSymbol.meta.implements = extractImplements(node);
    nodeSymbol.meta.type = "ClassDoc";
  } else if (types.isFunctionDeclaration(node) || types.isTSDeclareFunction(node)) {
    name = node.id ? node.id.name : "";
    nodeSymbol.meta.type = "FunctionDoc";
    nodeSymbol.meta.params = extractParams(node);
    nodeSymbol.meta.returns = extractReturns(node);
  } else if (types.isVariableDeclarator(node) || types.isObjectProperty(node) || types.isExpressionStatement(node) && types.isMemberExpression(node.expression.left) && !node.expression.left.computed) {
    let shouldSkip = false;

    if (types.isExpressionStatement(node)) {
      name = node.expression.left.property.name;
      init = node.expression.right;
      const rootName = resolveRootObject(node.expression.left);

      if (queryType(rootName) === PARAMETER) {
        shouldSkip = true;
        name = null;
      }
    } else if (types.isVariableDeclarator(node)) {
      name = node.id.name;
      init = resolveInit(node);

      if (nodePath.parent && nodePath.parent.kind === "const") {
        nodeSymbol.meta.readonly = true;
      }
    } else {
      name = types.isStringLiteral(node.key) ? node.key.value : node.key.name;
      init = node.value;
    }

    if (!shouldSkip) {
      if (types.isClassExpression(init) || types.isFunctionExpression(init) || types.isArrowFunctionExpression(init) || types.isCallExpression(init) && (types.isFunctionExpression(init.callee) || types.isArrowFunctionExpression(init.callee))) {
        flags |= languageLibrary.PASS_THROUGH | languageLibrary.VIRTUAL;
        isInit = true;
      } else if (!types.isObjectExpression(init)) {
        flags |= languageLibrary.OBLIGATE_LEAF;
        nodeSymbol.meta.type = "PropertyDoc";
      } else {
        nodeSymbol.meta.type = "PropertyDoc";
      }

      if (!isInit && init) {
        const [defaultValue, dataType] = extractAssignedValue(init);
        nodeSymbol.meta.defaultValue = defaultValue;
        nodeSymbol.meta.dataType = dataType;
      }

      let isExpression = false;

      if (types.isExpressionStatement(node)) {
        isExpression = true;
        const object = resolveObject(node.expression.left);
        const scope = isInstancePropertyAssignment(node.expression.left) ? "instance" : "static";
        nodeSymbol.meta.object = object;
        nodeSymbol.meta.scope = scope;

        if (scope === "instance" && object.includes(".prototype")) {
          nodeSymbol.meta.object = object.replace(".prototype", "");
        }
      } else if (types.isObjectProperty(node)) {
        nodeSymbol.meta.scope = "static";
      }

      if (types.isFunctionExpression(init) || types.isArrowFunctionExpression(init)) {
        nodeSymbol.meta.type = types.isObjectProperty(node) || isExpression ? "MethodDoc" : "FunctionDoc";
        nodeSymbol.meta.params = extractParams(init);
        nodeSymbol.meta.returns = extractReturns(init);
      } else if (types.isClassExpression(init)) {
        nodeSymbol.meta.type = "ClassDoc";
      }
    }
  } else if (types.isInterfaceDeclaration(node) || types.isTSInterfaceDeclaration(node)) {
    name = node.id.name;
    nodeSymbol.meta.extends = extractExtends(node);
    nodeSymbol.meta.type = "InterfaceDoc";
  } else if (types.isTSEnumDeclaration(node)) {
    name = node.id.name;
    nodeSymbol.meta.type = "EnumDoc";
  } else if (types.isTSEnumMember(node)) {
    name = node.id.name ? node.id.name : node.id.value;
    nodeSymbol.meta.constant = true;
    nodeSymbol.meta.value = types.isLiteral(node) ? node.value : null;
    nodeSymbol.meta.type = "PropertyDoc";
    flags |= languageLibrary.OBLIGATE_LEAF;
  } else if (types.isTSTypeElement(node)) {
    name = node.key ? node.key.name : "";

    if (types.isTSMethodSignature(node)) {
      nodeSymbol.meta.params = extractParams(node);
      nodeSymbol.meta.returns = extractReturns(node);
      nodeSymbol.meta.type = "MethodDoc";
    } else if (types.isTSPropertySignature(node)) {
      nodeSymbol.meta.readonly = node.readonly;
      nodeSymbol.meta.dataType = extractType(node);
      nodeSymbol.meta.type = "PropertyDoc";
    }

    flags |= languageLibrary.OBLIGATE_LEAF;
  } else if (parent && languageLibrary.isVirtual(parent) && types.isCallExpression(node) && (types.isFunctionExpression(node.callee) || types.isArrowFunctionExpression(node.callee))) {
    flags |= languageLibrary.VIRTUAL;
    const callee = node.callee;
    const returnedSymbol = resolveReturn(callee);

    if (returnedSymbol) {
      isInit = true;
      initComment = "";
      name = returnedSymbol;
    }
  }

  out.name = name;
  out.flags = flags;
  out.isInit = isInit;
  out.initComment = initComment;
  out.init = init;
  out.nodeSymbol = nodeSymbol;
}

function resolveInit(node) {
  if (types.isAssignmentExpression(node.init)) {
    return resolveInit(node.init);
  }

  return node.init;
}

function resolveReturn(callee) {
  if (callee.body && callee.body.body) {
    const body = callee.body.body;
    const lastStatement = body[body.length - 1];

    if (types.isReturnStatement(lastStatement) && lastStatement.argument && lastStatement.argument.name) {
      return lastStatement.argument.name;
    }
  }
}

function resolveObject(expression) {
  let longname = "";
  expression = expression.object;

  if (types.isTSAsExpression(expression)) {
    expression = expression.expression;
  }

  if (types.isThisExpression(expression)) {
    return "this";
  }

  while (expression.object) {
    longname = expression.property.name + (longname ? "." + longname : "");
    expression = expression.object;
  }

  longname = (types.isThisExpression(expression) ? "this" : expression.name) + (longname ? "." : "") + longname;
  return longname;
}

function resolveRootObject(expression) {
  expression = expression.object;

  if (types.isTSAsExpression(expression)) {
    expression = expression.expression;
  }

  if (types.isThisExpression(expression)) {
    return "this";
  }

  while (expression.object) {
    expression = expression.object;
  }

  return types.isThisExpression(expression) ? "this" : expression.name;
}

function isInstancePropertyAssignment(expr) {
  if (types.isThisExpression(expr.object)) {
    return true;
  }

  if (types.isMemberExpression(expr.object)) {
    if (expr.object.property.name === "prototype") {
      return true;
    }

    return isInstancePropertyAssignment(expr.object);
  }

  return false;
}

const SymbolUtils = {
  addChild: languageLibrary.addChildSymbol,

  commentIndex(comments) {
    for (let i = comments.length - 1; i >= 0; i--) {
      if (comments[i].value.startsWith("*") || comments[i].value.startsWith("!")) {
        return i;
      }
    }

    return comments.length - 1;
  },

  logRecursive(sym, prefix = "") {
    languageLibrary.log.info("Debug", prefix + (sym.simpleName || "-no-name-") + " " + sym.flags + ` [${sym.meta && sym.meta.type ? sym.meta.type : "-no-meta-"}]`);

    for (let i = 0; i < sym.members.length; i++) {
      SymbolUtils.logRecursive(sym.members[i], prefix + "\t");
    }
  },

  createModuleSymbol() {
    return {
      node: null,
      simpleName: "File",
      flags: 0,
      path: [],
      comment: "",
      canonicalName: "",
      parent: null,
      members: [],
      loc: _createLoc(),
      meta: {}
    };
  },

  createHeadlessSymbol(comment, loc) {
    return {
      node: null,
      simpleName: "",
      canonicalName: "",
      flags: 0,
      comment,
      members: [],
      loc: _createLoc(loc),
      meta: {},
      path: []
    };
  }

};
const extraVistors = {
  VariableDeclaration(nodePath) {
    const node = (nodePath.node);

    if (node.leadingComments && node.declarations && node.declarations.length && !node.declarations[0].leadingComments) {
      node.declarations[0].leadingComments = node.leadingComments;
      node.leadingComments = [];
    }
  }

};
let fileName = "";
let sourceFile;
let reportUndocumented = false;
function buildSymbolTree(file, source, config, plugins) {
  const moduleSymbol = SymbolUtils.createModuleSymbol();
  let ast;
  fileName = source.path;
  sourceFile = source;
  reportUndocumented = config.reportUndocumented;
  clearRegistry();

  try {
    ast = parser__namespace.parse(file, {
      plugins,
      sourceType: "module",
      errorRecovery: true
    });
  } catch (e) {
    console.error(`Babel couldn't parse file in @webdoc/parser (${fileName})`);
    throw e;
  }

  const ancestorStack = [moduleSymbol];
  createBlock();
  traverse__default["default"](ast, {
    enter(nodePath) {
      createBlock();
      const node = nodePath.node;

      if (extraVistors.hasOwnProperty(node.type)) {
        extraVistors[node.type](nodePath);
      }

      if (node.leadingComments && (types.isExportNamedDeclaration(node) || types.isExportDefaultDeclaration(node)) && node.declaration) {
        const declaration = nodePath.node.declaration;
        declaration.leadingComments = declaration.leadingComments || [];
        declaration.leadingComments.push(...node.leadingComments);
        node.leadingComments = [];
      }

      const scope = ancestorStack[ancestorStack.length - 1];
      let idoc;

      try {
        idoc = captureSymbols(nodePath, scope);
      } catch (e) {
        console.error(ancestorStack);
        console.log(node);
        console.error(ancestorStack.map(symbol => symbol.simpleName + `@{${symbol.comment}}[${symbol.node ? symbol.node.type : "Headless"}]`));
        throw e;
      }

      if (idoc && languageLibrary.isObligateLeaf(idoc)) {
        nodePath.skip(node);
        return;
      }

      if (idoc) {
        if (!languageLibrary.isVirtual(idoc) && !idoc.node) {
          console.log(node);
          console.log(idoc);
          throw new Error("WTF");
        }

        ancestorStack.push(idoc);
      }
    },

    exit(nodePath) {
      removeBlock();
      const currentPardoc = ancestorStack[ancestorStack.length - 1];

      if (currentPardoc && currentPardoc.node === nodePath.node) {
        ancestorStack.pop();

        if (languageLibrary.isVirtual(currentPardoc)) {
          const parentPardoc = (currentPardoc.parent);
          parentPardoc.members.splice(parentPardoc.members.indexOf(currentPardoc), 1);

          for (let i = 0; i < currentPardoc.members.length; i++) {
            currentPardoc.members[i].parent = null;
            SymbolUtils.addChild(currentPardoc.members[i], parentPardoc);
          }
        }
      }
    }

  });

  if (ancestorStack.length > 1) {
    console.error(ancestorStack);
    console.error(ancestorStack.map(symbol => symbol.simpleName + `@{${symbol.comment}}[${symbol.node ? symbol.node.type : "Headless"}]`));
    throw new Error("@webdoc/parser failed to correctly finish the symbol-tree.");
  }

  return moduleSymbol;
}

function captureSymbols(nodePath, parent) {
  const node = nodePath.node;
  const symbolInfo = {};
  extractSymbol(nodePath, parent, symbolInfo);
  let {
    name: simpleName,
    flags,
    init,
    initComment,
    isInit,
    nodeSymbol
  } = symbolInfo;

  if (nodeSymbol.meta.params) {
    const params = nodeSymbol.meta.params;

    for (let i = 0, j = params.length; i < j; i++) {
      declareParameter(params[i].identifier);
    }

    if (params.flawed) {
      console.error(languageLibrary.tag.Indexer, "^^^ Errors in parsing a parameter near " + fileName + ":" + (node && node.start ? node.start.line : "?"));
    }
  } else if (node.params) {
    registerParameters(node);
  }

  if (types.isVariableDeclaration(node)) {
    registerDeclaredVariables(node);
  }

  let nodeDocIndex;
  const leadingComments = filterComments(node.leadingComments) || [];

  if (leadingComments.length || types.isScope(node) || types.isTSParameterProperty(node) || reportUndocumented) {
    if (!initComment && leadingComments.length && simpleName) {
      nodeDocIndex = SymbolUtils.commentIndex(leadingComments);
    }

    const nodeDoc = typeof nodeDocIndex === "number" ? leadingComments[nodeDocIndex] : null;
    const comment = (initComment ? initComment : nodeDoc ? extract(nodeDoc) : "") || "";

    if (!comment) {
      nodeSymbol.meta.undocumented = true;

      if (types.isTSParameterProperty(node)) {
        nodeSymbol.meta.undocumentedAnchored = true;
      }
    }

    if (flags & languageLibrary.VIRTUAL && comment.indexOf("@member ") >= 0) {
      flags &= ~languageLibrary.VIRTUAL;
    }

    if (simpleName && (flags & languageLibrary.VIRTUAL) === 0) {
      nodeSymbol = Object.assign({
        node,
        simpleName,
        flags,
        path: [...parent.path, simpleName],
        comment,
        members: [],
        loc: _createLoc(nodeDoc ? nodeDoc.loc : node.loc ? node.loc : null)
      }, nodeSymbol);
      nodeSymbol = SymbolUtils.addChild(nodeSymbol, parent);
    } else if (simpleName && isInit) {
      nodeSymbol = Object.assign({
        node,
        simpleName: "",
        canonicalName: parent.canonicalName,
        flags,
        comment: "",
        path: [...parent.path],
        members: [{
          node: init,
          simpleName,
          flags: 0,
          comment,
          members: [],
          canonicalName: parent.canonicalName + "." + simpleName,
          path: [...parent.path, simpleName],
          loc: _createLoc(nodeDoc ? nodeDoc.loc : null),
          options: {
            object: node.expression ? node.expression.left.object.name : undefined
          },
          meta: { ...nodeSymbol.meta
          },
          __INITOR__: true
        }],
        loc: _createLoc(nodeDoc ? nodeDoc.loc : null),
        virtual: true
      }, nodeSymbol);
      nodeSymbol = SymbolUtils.addChild(nodeSymbol, parent);
      nodeSymbol.__INIT__ = true;
    } else {
      nodeSymbol = null;
    }

    for (let i = 0; i < leadingComments.length; i++) {
      if (i === nodeDocIndex) {
        continue;
      }

      const comment = extract(leadingComments[i]);

      if (!comment) {
        continue;
      }

      SymbolUtils.addChild(SymbolUtils.createHeadlessSymbol(comment, leadingComments[i].loc), parent);
    }
  } else {
    nodeSymbol = undefined;
  }

  const innerComments = node.innerComments;
  const trailingComments = node.trailingComments;

  if (nodeSymbol && innerComments) {
    const ns = nodeSymbol;
    innerComments.forEach(comment => {
      const doc = extract(comment);

      if (doc) {
        SymbolUtils.addChild(SymbolUtils.createHeadlessSymbol(doc, comment.loc), ns);
      }
    });
  }

  if (trailingComments) {
    const is11469 = types.isBlockStatement(node) && node.body.length === 0;

    const isTrailing = (n) => n.loc.end.line >= node.loc.end.line;

    trailingComments.forEach(comment => {
      const doc = extract(comment);

      if (doc) {
        const actualParent = is11469 && isTrailing(comment) ? parent.parent : parent;
        const tsym = SymbolUtils.createHeadlessSymbol(doc, comment.loc);
        SymbolUtils.addChild(tsym, actualParent);
      }
    });
  }

  return nodeSymbol;
}

const EMPTY_ARRAY = [];

function filterComments(comments) {
  return comments ? comments.filter(cnode => {
    return cnode.type === "CommentBlock" && (cnode.value.startsWith("*") || cnode.value.startsWith("!"));
  }) : EMPTY_ARRAY;
}

function registerDeclaredVariables(node) {
  const declarations = node.declarations;

  for (let i = 0, j = declarations.length; i < j; i++) {
    const decl = declarations[i];

    if (types.isIdentifier(decl.id)) {
      declareVariable(declarations[i].id.name);
    } else if (types.isObjectPattern(decl.id)) {
      registerObjectPropertyVariables(decl.id);
    } else if (types.isArrayPattern(decl.id)) {
      registerArrayElementVariables(decl.id);
    }
  }
}

function registerObjectPropertyVariables(node) {
  const props = node.properties;

  for (let i = 0, j = props.length; i < j; i++) {
    const prop = props[i];

    if (types.isRestElement(prop)) {
      continue;
    }

    if (!prop.key) {
      throw new Error("WTF");
    }

    declareVariable(prop.key.name);

    if (types.isObjectPattern(prop.value)) {
      registerObjectPropertyVariables(prop.value);
    } else if (types.isArrayPattern(prop.value)) {
      registerArrayElementVariables(prop.value);
    }
  }
}

function registerArrayElementVariables(node) {
  const elems = node.elements;

  for (let i = 0, j = elems.length; i < j; i++) {
    const elem = elems[i];

    if (types.isIdentifier(elem)) {
      declareVariable(elem.name);
    } else if (types.isObjectPattern(elem)) {
      registerObjectPropertyVariables(elem);
    } else if (types.isArrayPattern(elem)) {
      registerArrayElementVariables(elem);
    }
  }
}

function registerParameters(node) {
  if (!types.isFunctionExpression(node) && !types.isArrowFunctionExpression(node) && !types.isFunctionDeclaration(node) && !types.isClassMethod(node) && !types.isObjectMethod(node)) {
    return;
  }

  const params = extractParams(node);

  for (let i = 0, j = params.length; i < j; i++) {
    declareParameter(params[i].identifier);
  }
}

function _createLoc(loc) {
  if (!loc) {
    return {
      fileName,
      file: sourceFile,
      start: {
        line: NaN,
        column: NaN
      },
      end: {
        line: NaN,
        column: NaN
      }
    };
  }

  return {
    fileName,
    file: sourceFile,
    start: {
      line: loc.start.line,
      column: loc.start.column
    },
    end: {
      line: loc.end.line,
      column: loc.end.column
    }
  };
}

const defaultPreset = ["asyncGenerators", "bigInt", "classPrivateMethods", "classPrivateProperties", "classProperties", ["decorators", {
  decoratorsBeforeExport: true
}], "doExpressions", "dynamicImport", "exportDefaultFrom", "functionBind", "functionSent", "importMeta", "jsx", "logicalAssignment", "nullishCoalescingOperator", "numericSeparator", "objectRestSpread", "optionalCatchBinding", "optionalChaining", "throwExpressions"];
const flowPreset = [...defaultPreset, "flow", "flowComments"];
const tsPreset = [...defaultPreset, "typescript"];
const langJS = {
  extensions: ["js", "jsx", "jsdoc"],
  module: "@webdoc/language-babel/Language-JavaScript.js",

  parse(file, source, config) {
    mode.current = "flow";
    return buildSymbolTree(file, source, config, flowPreset);
  }

};
const langTS = {
  extensions: ["ts", "tsx"],
  module: "@webdoc/language-babel/Language-TypeScript.js",

  parse(file, source, config) {
    mode.current = "typescript";
    return buildSymbolTree(file, source, config, tsPreset);
  }

};

exports.langJS = langJS;
exports.langTS = langTS;
